import os
import re
from unittest.mock import patch

import pytest
import responses

from internetarchive.exceptions import DirectoryTraversalError
from internetarchive.utils import sanitize_filename
from tests.conftest import PROTOCOL, IaRequestsMock

DOWNLOAD_URL_RE = re.compile(f'{PROTOCOL}//archive.org/download/.*')
EXPECTED_LAST_MOD_HEADER = {"Last-Modified": "Tue, 14 Nov 2023 20:25:48 GMT"}


def test_file_download_sanitizes_filename(tmpdir, nasa_item):
    tmpdir.chdir()

    # Mock is_windows to return True to test Windows-style sanitization
    with patch('internetarchive.utils.is_windows', return_value=True):
        with IaRequestsMock(assert_all_requests_are_fired=False) as rsps:
            rsps.add(responses.GET, DOWNLOAD_URL_RE,
                     body='test content',
                     adding_headers=EXPECTED_LAST_MOD_HEADER)
            # Test filename with Windows-invalid characters
            file_obj = nasa_item.get_file('nasa_meta.xml')
            problematic_name = 'file:with<illegal>chars.xml'
            sanitized_name = sanitize_filename(problematic_name)
            expected_path = os.path.join(str(tmpdir), sanitized_name)

            file_obj.download(file_path=sanitized_name, destdir=str(tmpdir))
            assert os.path.exists(expected_path)


def test_file_download_prevents_directory_traversal(tmpdir, nasa_item):
    tmpdir.chdir()
    # Don't mock the request since it won't be made due to the security check
    with IaRequestsMock(assert_all_requests_are_fired=False):
        # Test directory traversal attempt by getting the file and calling download directly
        file_obj = nasa_item.get_file('nasa_meta.xml')
        malicious_path = os.path.join('..', 'nasa_meta.xml')
        with pytest.raises(DirectoryTraversalError, match="outside.*directory"):
            file_obj.download(file_path=malicious_path, destdir=str(tmpdir))



import string
import warnings
from unittest.mock import patch

import pytest

import internetarchive.utils
from tests.conftest import NASA_METADATA_PATH, IaRequestsMock


def test_utils():
    with open(__file__, encoding='utf-8') as fh:
        list(internetarchive.utils.chunk_generator(fh, 10))

    ifp = internetarchive.utils.IterableToFileAdapter([1, 2], 200)
    assert len(ifp) == 200
    ifp.read()


def test_needs_quote():
    notascii = ('ȧƈƈḗƞŧḗḓ ŧḗẋŧ ƒǿř ŧḗşŧīƞɠ, ℛℯα∂α♭ℓℯ ♭ʊ☂ η☺т Ѧ$☾ℐℐ, '
                '¡ooʇ ןnɟǝsn sı uʍop-ǝpısdn')
    assert internetarchive.utils.needs_quote(notascii)
    assert internetarchive.utils.needs_quote(string.whitespace)
    assert not internetarchive.utils.needs_quote(string.ascii_letters + string.digits)


def test_validate_s3_identifier():
    id1 = 'valid-Id-123-_foo'
    id2 = '!invalid-Id-123-_foo'
    id3 = 'invalid-Id-123-_foo+bar'
    id4 = 'invalid-Id-123-_føø'
    id5 = 'i'

    valid = internetarchive.utils.validate_s3_identifier(id1)
    assert valid

    for invalid_id in [id2, id3, id4, id5]:
        try:
            internetarchive.utils.validate_s3_identifier(invalid_id)
        except Exception as exc:
            assert isinstance(exc, internetarchive.utils.InvalidIdentifierException)


def test_get_md5():
    with open(__file__, 'rb') as fp:
        md5 = internetarchive.utils.get_md5(fp)
    assert isinstance(md5, str)


def test_IdentifierListAsItems(session):
    with IaRequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add_metadata_mock('nasa')
        it = internetarchive.utils.IdentifierListAsItems('nasa', session)
        assert it[0].identifier == 'nasa'
        assert it.nasa.identifier == 'nasa'


def test_IdentifierListAsItems_len(session):
    assert len(internetarchive.utils.IdentifierListAsItems(['foo', 'bar'], session)) == 2

# TODO: Add test of slice access to IdenfierListAsItems


def test_get_s3_xml_text():
    xml_str = ('<Error><Code>NoSuchBucket</Code>'
               '<Message>The specified bucket does not exist.</Message>'
               '<Resource>'
               'does-not-exist-! not found by Metadata::get_obj()[server]'
               '</Resource>'
               '<RequestId>d56bdc63-169b-4b4f-8c47-0fac6de39040</RequestId></Error>')

    expected_txt = internetarchive.utils.get_s3_xml_text(xml_str)
    assert expected_txt == ('The specified bucket does not exist. - does-not-exist-! '
                            'not found by Metadata::get_obj()[server]')


def test_get_file_size():
    try:
        s = internetarchive.utils.get_file_size(NASA_METADATA_PATH)
    except AttributeError as exc:
        assert "object has no attribute 'seek'" in str(exc)
    with open(NASA_METADATA_PATH) as fp:
        s = internetarchive.utils.get_file_size(fp)
    assert s == 7557


def test_is_valid_metadata_key():
    # Keys starting with "xml" should also be invalid
    # due to the XML specification, but are supported
    # by the Internet Archive.
    valid = ('adaptive_ocr', 'bookreader-defaults', 'frames_per_second',
             'identifier', 'possible-copyright-status', 'index[0]')
    invalid = ('Analog Format', "Date of transfer (probably today's date)",
               '_metadata_key', '58', '_', '<invalid>', 'a')

    for metadata_key in valid:
        assert internetarchive.utils.is_valid_metadata_key(metadata_key)

    for metadata_key in invalid:
        assert not internetarchive.utils.is_valid_metadata_key(metadata_key)


def test_is_windows():
    with patch('platform.system', return_value='Windows'), \
         patch('sys.platform', 'win32'):
        assert internetarchive.utils.is_windows() is True

    with patch('platform.system', return_value='Linux'), \
         patch('sys.platform', 'linux'):
        assert internetarchive.utils.is_windows() is False

def test_sanitize_filename_windows():
    test_cases = [
        ('file:name.txt', 'file%3Aname.txt'),
        ('file%name.txt', 'file%25name.txt'),
        ('con.txt', 'con.txt'),  # Reserved name, but no invalid chars so unchanged
        ('file .txt', 'file .txt'),  # Internal space preserved (not trailing)
        ('file  ', 'file'),  # Trailing spaces removed
        ('file..', 'file'),  # Trailing dots removed
        ('file . ', 'file'),  # Trailing space and dot removed
    ]

    for input_name, expected in test_cases:
        result = internetarchive.utils.sanitize_filename_windows(input_name)
        assert result == expected


def test_sanitize_filename_posix():
    # Test without colon encoding
    result = internetarchive.utils.sanitize_filename_posix('file/name.txt', False)
    assert result == 'file%2Fname.txt'

    # Test with colon encoding
    result = internetarchive.utils.sanitize_filename_posix('file:name.txt', True)
    assert result == 'file%3Aname.txt'

    # Test mixed encoding
    result = internetarchive.utils.sanitize_filename_posix('file/:name.txt', True)
    assert result == 'file%2F%3Aname.txt'


def test_unsanitize_filename():
    test_cases = [
        ('file%3Aname.txt', 'file:name.txt'),
        ('file%2Fname.txt', 'file/name.txt'),
        ('file%25name.txt', 'file%name.txt'),  # Percent sign
        ('normal.txt', 'normal.txt'),  # No encoding
    ]

    for input_name, expected in test_cases:
        with warnings.catch_warnings(record=True) as w:
            result = internetarchive.utils.unsanitize_filename(input_name)
            assert result == expected
            if '%' in input_name:
                assert len(w) == 1
                assert issubclass(w[0].category, UserWarning)


def test_sanitize_filename():
    # Test Windows path
    with patch('internetarchive.utils.is_windows', return_value=True):
        with warnings.catch_warnings(record=True) as w:
            result = internetarchive.utils.sanitize_filename('file:name.txt')
            assert result == 'file%3Aname.txt'
            assert len(w) == 1
            assert "sanitized" in str(w[0].message)

    # Test POSIX path
    with patch('internetarchive.utils.is_windows', return_value=False):
        result = internetarchive.utils.sanitize_filename('file/name.txt', False)
        assert result == 'file%2Fname.txt'


def test_sanitize_filepath():
    # Test with colon encoding
    result = internetarchive.utils.sanitize_filepath('/path/to/file:name.txt', True)
    assert result == '/path/to/file%3Aname.txt'

    # Test without colon encoding
    result = internetarchive.utils.sanitize_filepath('/path/to/file:name.txt', False)
    assert result == '/path/to/file:name.txt'  # Colon not encoded on POSIX by default

    # Test Windows path (mocked)
    with patch('internetarchive.utils.is_windows', return_value=True):
        result = internetarchive.utils.sanitize_filepath('/path/to/con.txt')
        assert result == '/path/to/con.txt'  # Reserved name sanitized



import os
import sys

import pytest

from internetarchive import get_item
from internetarchive.exceptions import DirectoryTraversalError
from internetarchive.files import File
from internetarchive.item import Item
from internetarchive.utils import (
    is_path_within_directory,
    sanitize_windows_filename,
    sanitize_windows_relpath,
)

IS_WIN = os.name == 'nt'

pytestmark = pytest.mark.skipif(not IS_WIN, reason='Windows specific tests')

def test_control_char_encoding():
    name = 'bad\x05name'
    sanitized, modified = sanitize_windows_filename(name)
    assert modified
    assert sanitized == 'bad%05name'

@pytest.mark.parametrize(('reserved','expected'), [
    ('AUX', 'AU%58'),
    ('CON', 'CO%4E'),
    ('COM1', 'COM%31'),
    ('LPT9', 'LPT%39'),
    ('NUL', 'NU%4C'),
])
def test_reserved_names(reserved, expected):
    sanitized, modified = sanitize_windows_filename(reserved)
    assert modified
    assert sanitized == expected

@pytest.mark.parametrize(('filename','expected'), [
    ('AUX.txt', 'AU%58.txt'),
    ('con.log', 'co%6E.log'),
    ('Com1.bin', 'Com%31.bin'),
    ('COM3.txt.txt', 'COM%33.txt.txt'),
])
def test_reserved_with_extension_sanitized(filename, expected):
    sanitized, modified = sanitize_windows_filename(filename)
    assert modified
    assert sanitized == expected

@pytest.mark.parametrize(('filename','expected'), [
    ('name.', 'name%2E'),
    ('name..', 'name%2E%2E'),
    ('trailspace ', 'trailspace%20'),
    ('both. ', 'both%2E%20'),
])
def test_trailing_dot_space(filename, expected):
    sanitized, modified = sanitize_windows_filename(filename)
    assert modified
    assert sanitized == expected

@pytest.mark.parametrize(('ch','enc'), [
    (':', '%3A'),
    ('*', '%2A'),
    ('<', '%3C'),
    ('>', '%3E'),
    ('|', '%7C'),
    ('?', '%3F'),
    ('\\', '%5C'),
    ('"', '%22')
])
def test_invalid_chars(ch, enc):
    sanitized, modified = sanitize_windows_filename(f'a{ch}b')
    assert modified
    assert sanitized == f'a{enc}b'

@pytest.mark.parametrize('name', [
    'back\\slash', 'dir\\\\file'
])
def test_backslash_always_encoded(name):
    sanitized, modified = sanitize_windows_filename(name)
    assert '%5C' in sanitized


def test_full_filename_combined_sanitization(tmp_path, monkeypatch):
    """Simulate downloading a file whose remote name contains many invalid characters
    including a backslash. We only test the sanitization logic up to path formation
    (not actual network download)."""
    remote_name = 'hello < > : " \\ | ? *.txt'
    # Use direct sanitize to assert expected output
    sanitized, modified = sanitize_windows_filename(remote_name)
    assert modified
    # Ensure each invalid char encoded
    for ch in ['<','>','|','?','*',':','\\','"',' ']:
        assert ch not in sanitized or ch == ' '  # trailing/inner spaces become %20
    assert '%5C' in sanitized  # backslash


def test_reserved_identifier_directory_sanitized(tmp_path):
    """Ensure that an item identifier that is a reserved device name is sanitized when
    constructing download paths."""
    # This test focuses on sanitize_windows_filename, as item.Download path building now
    # sanitizes components.
    reserved = 'AUX'
    sanitized, modified = sanitize_windows_filename(reserved)
    assert modified
    assert (sanitized.startswith('AU') and sanitized.endswith(b'X'.hex().upper()[:])) \
            or sanitized == 'AU%58'


def test_directory_traversal_exception_handled(monkeypatch, tmp_path):
    # Use is_path_within_directory directly
    base = tmp_path
    outside = tmp_path.parent / 'outside.txt'
    outside.write_text('x')
    assert not is_path_within_directory(str(base), str(outside))


@pytest.mark.parametrize('attempt', [
    '../evil.txt', '..\\evil.txt', '..%2Fevil.txt', '%2e%2e/evil.txt'
])
def test_traversal_attempt_sanitization(attempt):
    # sanitize_windows_relpath should NOT remove traversal but higher layer blocks it;
    # here we just ensure it encodes backslashes
    sanitized, _ = sanitize_windows_relpath(attempt, verbose=False)
    # Backslashes encoded
    if '\\' in attempt:
        assert '%5C' in sanitized or sanitized.replace('\\', '%5C')

@pytest.mark.parametrize('name', [
    'hello%20world', '%41already'
])
def test_existing_percent_sequences(name):
    # If no other encoding needed, percent remains unless part of %HH sequence
    # and no other changes?
    sanitized, modified = sanitize_windows_filename(name)
    # existing sequences remain unchanged because no other encoding triggered
    assert sanitized == name

@pytest.mark.parametrize('name', [
    'needs:encoding%20plus', 'AUX%41'  # reserved triggers change
])
def test_percent_gets_encoded_when_other_modifications(name):
    sanitized, modified = sanitize_windows_filename(name)
    if '%' in name and modified:
        assert '%25' in sanitized or name.count('%') == sanitized.count('%25')

# Directory traversal guard logic tests
# (cross-platform semantics validated on Windows here)

def test_is_path_within_directory_true(tmp_path):
    base = tmp_path
    target = base / 'subdir' / 'file.txt'
    target.parent.mkdir()
    target.write_text('x')
    assert is_path_within_directory(str(base), str(target))


def test_is_path_within_directory_false(tmp_path):
    base = tmp_path / 'a'
    other = tmp_path / 'b' / 'file.txt'
    base.mkdir()
    (tmp_path / 'b').mkdir()
    other.write_text('x')
    assert not is_path_within_directory(str(base), str(other))
